from datetime import datetime, timezone
from logging import getLogger

logger = getLogger("s3_scheduler")


class UnableToMoveFileError(Exception):
    pass


def nowut() -> datetime:
    """
    Current time with UTC timezone. use it instead of datetime.now()
    """
    return datetime.now(tz=timezone.utc)


def add_utc_tz(to_this: datetime) -> datetime:
    """
    Add timezone of utc to a naive representation of datetime, if time tz already exists then do not change anything
    :param to_this: Add relevant tz to this object.
    :return: Either new datetime if tz was added or the same that was given in param if no tz was added.
    """
    return to_this if to_this.tzinfo else to_this.replace(tzinfo=timezone.utc)


def move_file(source_bucket: str, target_bucket: str, source_file: str, s3_client, target_file: str = None):
    """
    Move a file from one bucket to the other.
    :param source_bucket:  Move a file from this bucket
    :param target_bucket: Move a file to this bucket
    :param source_file: File that exists in ;source_bucket' that you wish to move.
    :param s3_client: A boto S3 client used to interact with AWS api.
    :param target_file: If None then the target file name will be exactly like the source file name.
    """
    if not target_file:
        target_file = source_file
    if source_bucket and source_file and target_bucket:
        s3_client.copy_object(Bucket=target_bucket, CopySource=f"{source_bucket}/{source_file}", Key=target_file)
        s3_client.delete_object(Bucket=source_bucket, Key=source_file)
    else:
        raise UnableToMoveFileError()


def info(msg, *args, **kwargs):
    logger.info(msg, *args, **kwargs)


def warn(msg, *args, **kwargs):
    logger.warning(msg, *args, **kwargs)



from datetime import timedelta
from unittest import TestCase
from unittest.mock import MagicMock, patch
from urllib import parse

from s3_scheduler.scheduler import Scheduler
from s3_scheduler.utils import nowut


class SchedulerTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.s3_client = MagicMock()
        self.s3_resource = MagicMock()

        self.scheduler = Scheduler("bucket", "/path", self.s3_resource, self.s3_client)

    def test_stop_error_occured_return_false(self):
        self.s3_client.delete_object.side_effect = Exception()
        self.assertFalse(self.scheduler.stop("key"))

    def test_schedule_file_saved_in_right_format(self):
        now = nowut()
        upload_to = self.scheduler.schedule(now, "s3-bucket", "s3_files-important+", "content")

        self.assertEqual(upload_to, f'/path/|{parse.quote(str(now))}|s3-bucket|{parse.quote("s3_files-important+")}')

    @patch("s3_scheduler.scheduler.move_file")
    def test_handle_a_valid_file_time_arrived_file_is_copied(self, move_file_mock):
        time = nowut() - timedelta(minutes=10)
        upload_to = self.scheduler.schedule(time, "s3-bucket", "s3_files-important+", "content")

        mock_bucket = MagicMock(name="test")
        mock_result = MagicMock()
        mock_result.key = upload_to

        mock_bucket.objects.filter.return_value = [mock_result]
        self.s3_resource.Bucket.return_value = mock_bucket

        self.scheduler.handle()

        move_file_mock.assert_called_once_with("bucket", "s3-bucket", upload_to, self.s3_client, "s3_files-important+")

    @patch("s3_scheduler.utils.move_file")
    def test_handle_a_valid_file_time_not_arrived_file_is_not_copied(self, move_file_mock):
        time = nowut() + timedelta(minutes=10)
        upload_to = self.scheduler.schedule(time, "s3-bucket", "s3_files-important+", "content")

        mock_bucket = MagicMock(name="test")
        mock_result = MagicMock()
        mock_result.key = upload_to

        mock_bucket.objects.filter.return_value = [mock_result]
        self.s3_resource.Bucket.return_value = mock_bucket

        self.scheduler.handle()

        move_file_mock.assert_not_called()

    @patch("s3_scheduler.utils.move_file")
    def test_handle_an_invalid_file_file_is_not_copied(self, move_file_mock):
        time = nowut() - timedelta(minutes=10)
        upload_to = self.scheduler.schedule(time, "s3-bucket", "s3_files-important+", "content")

        mock_bucket = MagicMock(name="test")
        mock_result = MagicMock()
        mock_result.key = upload_to + "|ddd"

        mock_bucket.objects.filter.return_value = [mock_result]
        self.s3_resource.Bucket.return_value = mock_bucket

        self.scheduler.handle()

        move_file_mock.assert_not_called()

    @patch("s3_scheduler.utils.move_file")
    def test_handle_an_invalid_time_format_file_is_not_copied(self, move_file_mock):
        time = nowut() - timedelta(minutes=10)
        upload_to = self.scheduler.schedule(time, "s3-bucket", "s3_files-important+", "content")

        mock_bucket = MagicMock(name="test")
        mock_result = MagicMock()
        mock_result.key = upload_to.replace(str(time.year), "Hello")

        mock_bucket.objects.filter.return_value = [mock_result]
        self.s3_resource.Bucket.return_value = mock_bucket

        self.scheduler.handle()

        move_file_mock.assert_not_called()

    @patch("s3_scheduler.scheduler.move_file")
    @patch("s3_scheduler.scheduler.warn")
    def test_handle_unable_to_copy_file_handle_gracefully(self, warn_mock, move_file_mock):
        time = nowut() - timedelta(minutes=10)
        upload_to = self.scheduler.schedule(time, "s3-bucket", "s3_files-important+", "content")

        mock_bucket = MagicMock(name="test")
        mock_result = MagicMock()
        mock_result.key = upload_to

        mock_bucket.objects.filter.return_value = [mock_result]
        self.s3_resource.Bucket.return_value = mock_bucket
        move_file_mock.side_effect = Exception()

        # noinspection PyTypeChecker
        self.scheduler.handle()

        warn_mock.assert_called_once()



import io
from datetime import datetime

from dateutil.parser import parse as date_parser
from typing import Tuple, Any
from urllib import parse

from .utils import nowut, info, warn, move_file
from .utils import add_utc_tz


class Scheduler:
    def __init__(self, bucket: str, path: str, s3_resource, s3_client) -> None:
        """
        :param bucket: In which bucket to store the scheduled tasks
        :param path: The key in S3 to keep the scheduled tasks
        """
        super().__init__()
        self._bucket = bucket
        self._path = path if not path.endswith("/") else path[:-1]
        self._s3_resource = s3_resource

        self._s3_client = s3_client

    def schedule(self, when: datetime, bucket: str, path: str, what: str) -> str:
        """
        Schedule a file to be copied to a specific location.
        :param when: When to schedule the file copying
        :param bucket: Copy to this bucket
        :param path:Path to which to copy, includes the directory and file name. Should not start with '/'
        :param what: The actual content to copy.
        :return: Full path of where the file was copied to
        """
        when_escaped = parse.quote(str(add_utc_tz(when)))
        bucket_escaped = parse.quote(bucket)
        # / is treated as directory mark in S3 so replace it with something else
        path_escaped = parse.quote(path if not path.startswith("/") else path[1:]).replace("/", "~~")

        if "|" in when_escaped or "|" in bucket_escaped or "|" in path_escaped:
            raise ValueError(f'{when_escaped} or {bucket_escaped} or {path_escaped} has "|" in it')
        file_name = f"|{when_escaped}|{bucket_escaped}|{path_escaped}"
        fo = io.BytesIO(bytes(what, encoding="utf-8"))

        upload_to = f"{self._path}/{file_name}"
        self._s3_client.upload_fileobj(fo, self._bucket, upload_to)
        info(f"Scheduled on {what} to copy {path}")
        return upload_to

    def stop(self, key: Any) -> bool:
        """
        Stop a scheduled task donated by 'key'. Return True on success, false otherwise
        """
        # noinspection PyBroadException
        try:
            result: dict = self._s3_client.delete_object(Bucket=self._bucket, Key=key)
            info(f"Deleted a scheduled task {key}.")
            return result["ResponseMetadata"]["HTTPStatusCode"] in [204, 200]
        except Exception as ex:
            warn(f"Unable to delete key in scheduler - {ex}")
            return False

    def handle(self,):
        # Get a list of all files in an s3 path
        bucket = self._s3_resource.Bucket(self._bucket)

        # check if time has passed, if it did then copy the file to the relevant bucket
        requires_handling = []

        for obj in bucket.objects.filter(Prefix=self._path):
            if self._time_passed(obj.key):
                # noinspection PyBroadException
                try:
                    requires_handling.append(self._extract_location(obj.key))
                except Exception as ex:
                    warn(f"{obj} has invalid structure, ignoring scheduled task. {ex}")

        for source_file, target_bucket, target_path in requires_handling:
            # noinspection PyBroadException
            try:
                move_file(self._bucket, target_bucket, source_file, self._s3_client, target_path)
            except Exception as ex:
                warn(f"Unable to copy {source_file} to {bucket}/{target_path}. {ex}")

    def _time_passed(self, key: str) -> bool:
        elements = key.split("|")
        if len(elements) != 4:
            warn(f"{key} has invalid structure, ignoring scheduled task")
            return False

        try:
            time = date_parser(parse.unquote(elements[1]))
            return nowut() > time
        except ValueError:
            warn(f"{key} has invalid structure, ignoring scheduled task")
            return False

    def _extract_location(self, key: str) -> Tuple[str, str, str]:
        """
        Extract the components from the key's name
        :param key: Key to extract from
        :return: A tuple of original's key name, target  buck, target path
        """
        elements = key.split("|")

        bucket = parse.unquote(elements[2])
        target_path = parse.unquote(elements[3]).replace("~~", "/")

        return key, bucket, target_path
